@@ -18,18 +18,18 @@
 package org.apache.cassandra.db.rows;
 
 import java.util.*;
-import java.security.MessageDigest;
+import java.util.function.BiConsumer;
 import java.util.function.Consumer;
 
-import com.google.common.base.Predicate;
-
-import org.apache.cassandra.config.CFMetaData;
-import org.apache.cassandra.config.ColumnDefinition;
 import org.apache.cassandra.db.*;
 import org.apache.cassandra.db.filter.ColumnFilter;
+import org.apache.cassandra.schema.ColumnMetadata;
+import org.apache.cassandra.schema.TableMetadata;
 import org.apache.cassandra.service.paxos.Commit;
-import org.apache.cassandra.utils.FBUtilities;
+import org.apache.cassandra.utils.BiLongAccumulator;
+import org.apache.cassandra.utils.LongAccumulator;
 import org.apache.cassandra.utils.MergeIterator;
+import org.apache.cassandra.utils.ObjectSizes;
 import org.apache.cassandra.utils.SearchIterator;
 import org.apache.cassandra.utils.btree.BTree;
 import org.apache.cassandra.utils.btree.UpdateFunction;
@@ -54,13 +54,13 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
      * The clustering values for this row.
      */
     @Override
-    public Clustering clustering();
+    public Clustering<?> clustering();
 
     /**
      * An in-natural-order collection of the columns for which data (incl. simple tombstones)
      * is present in this row.
      */
-    public Collection<ColumnDefinition> columns();
+    public Collection<ColumnMetadata> columns();
 
 
     /**
@@ -126,7 +126,7 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
      * @param c the simple column for which to fetch the cell.
      * @return the corresponding cell or {@code null} if the row has no such cell.
      */
-    public Cell getCell(ColumnDefinition c);
+    public Cell<?> getCell(ColumnMetadata c);
 
     /**
      * Return a cell for a given complex column and cell path.
@@ -135,7 +135,7 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
      * @param path the cell path for which to fetch the cell.
      * @return the corresponding cell or {@code null} if the row has no such cell.
      */
-    public Cell getCell(ColumnDefinition c, CellPath path);
+    public Cell<?> getCell(ColumnMetadata c, CellPath path);
 
     /**
      * The data for a complex column.
@@ -145,7 +145,7 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
      * @param c the complex column for which to return the complex data.
      * @return the data for {@code c} or {@code null} if the row has no data for this column.
      */
-    public ComplexColumnData getComplexColumnData(ColumnDefinition c);
+    public ComplexColumnData getComplexColumnData(ColumnMetadata c);
 
     /**
      * An iterable over the cells of this row.
@@ -154,7 +154,7 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
      *
      * @return an iterable over the cells of this row.
      */
-    public Iterable<Cell> cells();
+    public Iterable<Cell<?>> cells();
 
     /**
      * A collection of the ColumnData representation of this row, for columns with some data (possibly not live) present
@@ -176,7 +176,7 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
      * @param reversed if cells should returned in reverse order.
      * @return an iterable over the cells of this row in "legacy order".
      */
-    public Iterable<Cell> cellsInLegacyOrder(CFMetaData metadata, boolean reversed);
+    public Iterable<Cell<?>> cellsInLegacyOrder(TableMetadata metadata, boolean reversed);
 
     /**
      * Whether the row stores any (non-live) complex deletion for any complex column.
@@ -200,14 +200,14 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
      *
      * @return a search iterator for the cells of this row.
      */
-    public SearchIterator<ColumnDefinition, ColumnData> searchIterator();
+    public SearchIterator<ColumnMetadata, ColumnData> searchIterator();
 
     /**
      * Returns a copy of this row that:
      *   1) only includes the data for the column included by {@code filter}.
      *   2) doesn't include any data that belongs to a dropped column (recorded in {@code metadata}).
      */
-    public Row filter(ColumnFilter filter, CFMetaData metadata);
+    public Row filter(ColumnFilter filter, TableMetadata metadata);
 
     /**
      * Returns a copy of this row that:
@@ -216,7 +216,7 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
      *   3) doesn't include any data that is shadowed/deleted by {@code activeDeletion}.
      *   4) uses {@code activeDeletion} as row deletion iff {@code setActiveDeletionToRow} and {@code activeDeletion} supersedes the row deletion.
      */
-    public Row filter(ColumnFilter filter, DeletionTime activeDeletion, boolean setActiveDeletionToRow, CFMetaData metadata);
+    public Row filter(ColumnFilter filter, DeletionTime activeDeletion, boolean setActiveDeletionToRow, TableMetadata metadata);
 
     /**
      * Returns a copy of this row without any deletion info that should be purged according to {@code purger}.
@@ -224,7 +224,7 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
      * @param purger the {@code DeletionPurger} to use to decide what can be purged.
      * @param nowInSec the current time to decide what is deleted and what isn't (in the case of expired cells).
      * @param enforceStrictLiveness whether the row should be purged if there is no PK liveness info,
-     *                              normally retrieved from {@link CFMetaData#enforceStrictLiveness()}
+     *                              normally retrieved from {@link TableMetadata#enforceStrictLiveness()}
      *
      *        When enforceStrictLiveness is set, rows with empty PK liveness info
      *        and no row deletion are purged.
@@ -234,7 +234,7 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
      *        is not live. See CASSANDRA-11500.
      *
      * @return this row but without any deletion info purged by {@code purger}. If the purged row is empty, returns
-     * {@code null}.
+     *         {@code null}.
      */
     public Row purge(DeletionPurger purger, int nowInSec, boolean enforceStrictLiveness);
 
@@ -279,17 +279,29 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
 
     public long unsharedHeapSizeExcludingData();
 
-    public String toString(CFMetaData metadata, boolean fullDetails);
+    public String toString(TableMetadata metadata, boolean fullDetails);
 
     /**
      * Apply a function to every column in a row
      */
-    public void apply(Consumer<ColumnData> function, boolean reverse);
+    public void apply(Consumer<ColumnData> function);
 
     /**
-     * Apply a funtion to every column in a row until a stop condition is reached
+     * Apply a function to every column in a row
      */
-    public void apply(Consumer<ColumnData> function, Predicate<ColumnData> stopCondition, boolean reverse);
+    public <A> void apply(BiConsumer<A, ColumnData> function, A arg);
+
+    /**
+     * Apply an accumulation funtion to every column in a row
+     */
+
+    public long accumulate(LongAccumulator<ColumnData> accumulator, long initialValue);
+
+    public long accumulate(LongAccumulator<ColumnData> accumulator, Comparator<ColumnData> comparator, ColumnData from, long initialValue);
+
+    public <A> long accumulate(BiLongAccumulator<A, ColumnData> accumulator, A arg, long initialValue);
+
+    public <A> long accumulate(BiLongAccumulator<A, ColumnData> accumulator, A arg, Comparator<ColumnData> comparator, ColumnData from, long initialValue);
 
     /**
      * A row deletion/tombstone.
@@ -312,6 +324,7 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
     public static class Deletion
     {
         public static final Deletion LIVE = new Deletion(DeletionTime.LIVE, false);
+        private static final long EMPTY_SIZE = ObjectSizes.measure(new DeletionTime(0, 0));
 
         private final DeletionTime time;
         private final boolean isShadowable;
@@ -386,15 +399,15 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
             return time.deletes(info);
         }
 
-        public boolean deletes(Cell cell)
+        public boolean deletes(Cell<?> cell)
         {
             return time.deletes(cell);
         }
 
-        public void digest(MessageDigest digest)
+        public void digest(Digest digest)
         {
             time.digest(digest);
-            FBUtilities.updateWithBoolean(digest, isShadowable);
+            digest.updateWithBoolean(isShadowable);
         }
 
         public int dataSize()
@@ -411,6 +424,14 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
             return this.time.equals(that.time) && this.isShadowable == that.isShadowable;
         }
 
+        public long unsharedHeapSize()
+        {
+            if(this == LIVE)
+                return 0;
+
+            return EMPTY_SIZE + time().unsharedHeapSize();
+        }
+
         @Override
         public final int hashCode()
         {
@@ -471,7 +492,7 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
          *
          * @param clustering the clustering for the new row.
          */
-        public void newRow(Clustering clustering);
+        public void newRow(Clustering<?> clustering);
 
         /**
          * The clustering for the row that is currently being built.
@@ -479,7 +500,7 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
          * @return the clustering for the row that is currently being built, or {@code null} if {@link #newRow} hasn't
          * yet been called.
          */
-        public Clustering clustering();
+        public Clustering<?> clustering();
 
         /**
          * Adds the liveness information for the partition key columns of this row.
@@ -504,7 +525,7 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
          *
          * @param cell the cell to add.
          */
-        public void addCell(Cell cell);
+        public void addCell(Cell<?> cell);
 
         /**
          * Adds a complex deletion.
@@ -512,7 +533,7 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
          * @param column the column for which to add the {@code complexDeletion}.
          * @param complexDeletion the complex deletion time to add.
          */
-        public void addComplexDeletion(ColumnDefinition column, DeletionTime complexDeletion);
+        public void addComplexDeletion(ColumnMetadata column, DeletionTime complexDeletion);
 
         /**
          * Builds and return built row.
@@ -629,18 +650,18 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
         private final Row[] rows;
         private final List<Iterator<ColumnData>> columnDataIterators;
 
-        private Clustering clustering;
+        private Clustering<?> clustering;
         private int rowsToMerge;
         private int lastRowSet = -1;
 
         private final List<ColumnData> dataBuffer = new ArrayList<>();
         private final ColumnDataReducer columnDataReducer;
 
-        public Merger(int size, int nowInSec, boolean hasComplex)
+        public Merger(int size, boolean hasComplex)
         {
             this.rows = new Row[size];
             this.columnDataIterators = new ArrayList<>(size);
-            this.columnDataReducer = new ColumnDataReducer(size, nowInSec, hasComplex);
+            this.columnDataReducer = new ColumnDataReducer(size, hasComplex);
         }
 
         public void clear()
@@ -660,6 +681,7 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
             lastRowSet = i;
         }
 
+        @SuppressWarnings("resource")
         public Row merge(DeletionTime activeDeletion)
         {
             // If for this clustering we have only one row version and have no activeDeletion (i.e. nothing to filter out),
@@ -710,10 +732,10 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
             // Because some data might have been shadowed by the 'activeDeletion', we could have an empty row
             return rowInfo.isEmpty() && rowDeletion.isLive() && dataBuffer.isEmpty()
                  ? null
-                 : BTreeRow.create(clustering, rowInfo, rowDeletion, BTree.build(dataBuffer, UpdateFunction.<ColumnData>noOp()));
+                 : BTreeRow.create(clustering, rowInfo, rowDeletion, BTree.build(dataBuffer));
         }
 
-        public Clustering mergedClustering()
+        public Clustering<?> mergedClustering()
         {
             return clustering;
         }
@@ -725,24 +747,21 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
 
         private static class ColumnDataReducer extends MergeIterator.Reducer<ColumnData, ColumnData>
         {
-            private final int nowInSec;
-
-            private ColumnDefinition column;
+            private ColumnMetadata column;
             private final List<ColumnData> versions;
 
             private DeletionTime activeDeletion;
 
             private final ComplexColumnData.Builder complexBuilder;
-            private final List<Iterator<Cell>> complexCells;
+            private final List<Iterator<Cell<?>>> complexCells;
             private final CellReducer cellReducer;
 
-            public ColumnDataReducer(int size, int nowInSec, boolean hasComplex)
+            public ColumnDataReducer(int size, boolean hasComplex)
             {
-                this.nowInSec = nowInSec;
                 this.versions = new ArrayList<>(size);
                 this.complexBuilder = hasComplex ? ComplexColumnData.builder() : null;
                 this.complexCells = hasComplex ? new ArrayList<>(size) : null;
-                this.cellReducer = new CellReducer(nowInSec);
+                this.cellReducer = new CellReducer();
             }
 
             public void setActiveDeletion(DeletionTime activeDeletion)
@@ -752,35 +771,36 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
 
             public void reduce(int idx, ColumnData data)
             {
-                if (useColumnDefinition(data.column()))
+                if (useColumnMetadata(data.column()))
                     column = data.column();
 
                 versions.add(data);
             }
 
             /**
-             * Determines it the {@code ColumnDefinition} is the one that should be used.
-             * @param dataColumn the {@code ColumnDefinition} to use.
-             * @return {@code true} if the {@code ColumnDefinition} is the one that should be used, {@code false} otherwise.
+             * Determines it the {@code ColumnMetadata} is the one that should be used.
+             * @param dataColumn the {@code ColumnMetadata} to use.
+             * @return {@code true} if the {@code ColumnMetadata} is the one that should be used, {@code false} otherwise.
              */
-            private boolean useColumnDefinition(ColumnDefinition dataColumn)
+            private boolean useColumnMetadata(ColumnMetadata dataColumn)
             {
                 if (column == null)
                     return true;
 
-                return ColumnDefinitionVersionComparator.INSTANCE.compare(column, dataColumn) < 0;
+                return ColumnMetadataVersionComparator.INSTANCE.compare(column, dataColumn) < 0;
             }
 
+            @SuppressWarnings("resource")
             protected ColumnData getReduced()
             {
                 if (column.isSimple())
                 {
-                    Cell merged = null;
-                    for (ColumnData data : versions)
+                    Cell<?> merged = null;
+                    for (int i=0, isize=versions.size(); i<isize; i++)
                     {
-                        Cell cell = (Cell)data;
+                        Cell<?> cell = (Cell<?>) versions.get(i);
                         if (!activeDeletion.deletes(cell))
-                            merged = merged == null ? cell : Cells.reconcile(merged, cell, nowInSec);
+                            merged = merged == null ? cell : Cells.reconcile(merged, cell);
                     }
                     return merged;
                 }
@@ -789,8 +809,9 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
                     complexBuilder.newColumn(column);
                     complexCells.clear();
                     DeletionTime complexDeletion = DeletionTime.LIVE;
-                    for (ColumnData data : versions)
+                    for (int i=0, isize=versions.size(); i<isize; i++)
                     {
+                        ColumnData data = versions.get(i);
                         ComplexColumnData cd = (ComplexColumnData)data;
                         if (cd.complexDeletion().supersedes(complexDeletion))
                             complexDeletion = cd.complexDeletion();
@@ -807,10 +828,10 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
                         cellReducer.setActiveDeletion(activeDeletion);
                     }
 
-                    Iterator<Cell> cells = MergeIterator.get(complexCells, Cell.comparator, cellReducer);
+                    Iterator<Cell<?>> cells = MergeIterator.get(complexCells, Cell.comparator, cellReducer);
                     while (cells.hasNext())
                     {
-                        Cell merged = cells.next();
+                        Cell<?> merged = cells.next();
                         if (merged != null)
                             complexBuilder.addCell(merged);
                     }
@@ -825,17 +846,10 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
             }
         }
 
-        private static class CellReducer extends MergeIterator.Reducer<Cell, Cell>
+        private static class CellReducer extends MergeIterator.Reducer<Cell<?>, Cell<?>>
         {
-            private final int nowInSec;
-
             private DeletionTime activeDeletion;
-            private Cell merged;
-
-            public CellReducer(int nowInSec)
-            {
-                this.nowInSec = nowInSec;
-            }
+            private Cell<?> merged;
 
             public void setActiveDeletion(DeletionTime activeDeletion)
             {
@@ -843,13 +857,13 @@ public interface Row extends Unfiltered, Iterable<ColumnData>
                 onKeyChange();
             }
 
-            public void reduce(int idx, Cell cell)
+            public void reduce(int idx, Cell<?> cell)
             {
                 if (!activeDeletion.deletes(cell))
-                    merged = merged == null ? cell : Cells.reconcile(merged, cell, nowInSec);
+                    merged = merged == null ? cell : Cells.reconcile(merged, cell);
             }
 
-            protected Cell getReduced()
+            protected Cell<?> getReduced()
             {
                 return merged;
             }
