@@ -25,7 +25,7 @@ import java.util.Arrays;
 import java.util.Collections;
 import java.util.HashSet;
 import java.util.Set;
-import java.util.function.Function;
+import java.util.function.ToIntFunction;
 
 import com.google.common.collect.Iterables;
 import org.junit.Test;
@@ -71,17 +71,6 @@ public class GcCompactionTest extends CQLTester
 
     @Test
     public void testGcCompactionPartitions() throws Throwable
-    {
-        testGcCompactionPartitions(false);
-    }
-
-    @Test
-    public void testGcCompactionPartitionsUnrepaired() throws Throwable
-    {
-        testGcCompactionPartitions(true);
-    }
-
-    public void testGcCompactionPartitions(boolean onlyPurgeRepairedTombstones) throws Throwable
     {
         runCompactionTest("CREATE TABLE %s(" +
                           "  key int," +
@@ -89,25 +78,13 @@ public class GcCompactionTest extends CQLTester
                           "  data int," +
                           "  extra text," +
                           "  PRIMARY KEY((key, column), data)" +
-                          ") WITH compaction = { 'class' :  'SizeTieredCompactionStrategy', " +
-                          "'provide_overlapping_tombstones' : 'row', " +
-                          "'only_purge_repaired_tombstones': " + onlyPurgeRepairedTombstones + " };",
-                          onlyPurgeRepairedTombstones);
+                          ") WITH compaction = { 'class' :  'SizeTieredCompactionStrategy', 'provide_overlapping_tombstones' : 'row'  };"
+                          );
+
     }
 
     @Test
     public void testGcCompactionRows() throws Throwable
-    {
-        testGcCompactionRows(false);
-    }
-
-    @Test
-    public void testGcCompactionRowsUnrepaired() throws Throwable
-    {
-        testGcCompactionRows(true);
-    }
-
-    public void testGcCompactionRows(boolean onlyPurgeRepairedTombstones) throws Throwable
     {
         runCompactionTest("CREATE TABLE %s(" +
                           "  key int," +
@@ -115,27 +92,13 @@ public class GcCompactionTest extends CQLTester
                           "  data int," +
                           "  extra text," +
                           "  PRIMARY KEY(key, column)" +
-                          ") WITH compaction = { 'class' :  'SizeTieredCompactionStrategy', " +
-                          "'provide_overlapping_tombstones' : 'row', " +
-                          "'only_purge_repaired_tombstones': " + onlyPurgeRepairedTombstones + " };",
-                          onlyPurgeRepairedTombstones
-        );
+                          ") WITH compaction = { 'class' :  'SizeTieredCompactionStrategy', 'provide_overlapping_tombstones' : 'row'  };"
+                          );
 
     }
 
     @Test
     public void testGcCompactionRanges() throws Throwable
-    {
-        testGcCompactionRanges(false);
-    }
-
-    @Test
-    public void testGcCompactionRangesUnrepaired() throws Throwable
-    {
-        testGcCompactionRanges(true);
-    }
-
-    public void testGcCompactionRanges(boolean onlyPurgeRepairedTombstones) throws Throwable
     {
 
         runCompactionTest("CREATE TABLE %s(" +
@@ -145,14 +108,11 @@ public class GcCompactionTest extends CQLTester
                           "  data int," +
                           "  extra text," +
                           "  PRIMARY KEY(key, column, data)" +
-                          ") WITH compaction = { 'class' :  'SizeTieredCompactionStrategy', " +
-                          "'provide_overlapping_tombstones' : 'row', " +
-                          "'only_purge_repaired_tombstones': " + onlyPurgeRepairedTombstones + " };",
-                          onlyPurgeRepairedTombstones
-        );
+                          ") WITH compaction = { 'class' :  'SizeTieredCompactionStrategy', 'provide_overlapping_tombstones' : 'row'  };"
+                          );
     }
 
-    private void runCompactionTest(String tableDef, boolean onlyPurgeRepairedTombstones) throws Throwable
+    private void runCompactionTest(String tableDef) throws Throwable
     {
         createTable(tableDef);
 
@@ -188,7 +148,7 @@ public class GcCompactionTest extends CQLTester
         assertEquals(3, cfs.getLiveSSTables().size());
         SSTableReader table3 = getNewTable(readers);
         assertEquals(0, countTombstoneMarkers(table3));
-        assertTrue(onlyPurgeRepairedTombstones ? rowCount == countRows(table3) : rowCount > countRows(table3));
+        assertTrue(rowCount > countRows(table3));
     }
 
     @Test
@@ -511,7 +471,7 @@ public class GcCompactionTest extends CQLTester
 
     int countRows(SSTableReader reader)
     {
-        boolean enforceStrictLiveness = reader.metadata.enforceStrictLiveness();
+        boolean enforceStrictLiveness = reader.metadata().enforceStrictLiveness();
         int nowInSec = FBUtilities.nowInSeconds();
         return count(reader, x -> x.isRow() && ((Row) x).hasLiveData(nowInSec, enforceStrictLiveness) ? 1 : 0, x -> 0);
     }
@@ -539,7 +499,7 @@ public class GcCompactionTest extends CQLTester
         return ccd.cellsCount();
     }
 
-    int count(SSTableReader reader, Function<Unfiltered, Integer> predicate, Function<UnfilteredRowIterator, Integer> partitionPredicate)
+    int count(SSTableReader reader, ToIntFunction<Unfiltered> predicate, ToIntFunction<UnfilteredRowIterator> partitionPredicate)
     {
         int instances = 0;
         try (ISSTableScanner partitions = reader.getScanner())
@@ -548,11 +508,11 @@ public class GcCompactionTest extends CQLTester
             {
                 try (UnfilteredRowIterator iter = partitions.next())
                 {
-                    instances += partitionPredicate.apply(iter);
+                    instances += partitionPredicate.applyAsInt(iter);
                     while (iter.hasNext())
                     {
                         Unfiltered atom = iter.next();
-                        instances += predicate.apply(atom);
+                        instances += predicate.applyAsInt(atom);
                     }
                 }
             }
